#ifndef CERF_DAWSON_H
#define CERF_DAWSON_H

/**
 * Author = robbert
 * Date = 2014-05-17
 * License = LGPL v3
 * Maintainer = Robbert Harms
 * Email = robbert.harms@maastrichtuniversity.nl
 */

 /**
 * Calculate the Dawson's integral for a real argument.
 */
double dawson(double x);

#endif // CERF_DAWSON_H
