#ifndef CERF_ERFI_H
#define CERF_ERFI_H

/**
 * Author = robbert
 * Date = 2014-05-17
 * License = LGPL v3
 * Maintainer = Robbert Harms
 * Email = robbert.harms@maastrichtuniversity.nl
 */

/**
 * Calculate the imaginary error function for a real argument (special case)
 */
double erfi(double x);
float ferfi(float x);

#endif //CERF_ERFI_H
