#ifndef CERF_IM_W_OF_X_H
#define CERF_IM_W_OF_X_H

/**
 * Author = robbert
 * Date = 2014-05-17
 * License = LGPL v3
 * Maintainer = Robbert Harms
 * Email = robbert.harms@maastrichtuniversity.nl
 */

double im_w_of_x(double x);

#endif // CERF_IM_W_OF_X_H
